#include <iostream>
#include <cmath>
#include <armadillo>
#include <boost/asio.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include "ros/ros.h"
#include "ros/console.h"
#include "udrone/desiredPosition.h"
#include "udrone/statusControlAlgorithms.h"
#include "path_following/update_path.h"
#include "path_following/desired_velocity.h"
#include <sstream>

using namespace arma;

class PathFollowing
{
public:
    PathFollowing(ros::NodeHandle nh){
        this->n = nh;

        pubConfirmChangeCtrlAlg     = nh.advertise<udrone::statusControlAlgorithms>("confirmChangeCtrlAlg",10);
        subChangeCtrlAlg            = nh.subscribe("changeCtrlAlg", 10, &PathFollowing::activate, this);

        this->pf_active                           = false;
        this->returnMsg.confirmationMsg           = true;
        this->returnMsg.statusDirectMotionControl = false;
        this->returnMsg.statusDirectThrustControl = false;
        this->returnMsg.statusAutoDepthControl    = false;
        this->returnMsg.statusAutoHeadingControl  = false;
        this->returnMsg.statusPathfollowing       = true;

        this->path_set = 0;
    }
    void updatePathCallback(const path_following::update_path& msg);
    void setVelocityCallback(const path_following::desired_velocity& msg);
    void updateDesired(float loop_periode);
    void activate(const udrone::statusControlAlgorithms);

    float lineareInterpolation(float x, float x0, float x1, float y0, float y1);

    float ds = 0.1;
    float S;

    float current_x_desired;
    float current_y_desired;
    float current_z_desired;
    float current_psi_desired;

    bool pf_active = false;

private:
    ros::NodeHandle n;
    ros::Publisher pubConfirmChangeCtrlAlg;
    ros::Subscriber subChangeCtrlAlg;

    udrone::statusControlAlgorithms returnMsg;

    int path_set;

    int Nl;
    float kappa;
    float desired_velocity;

    mat x_desired;
    mat y_desired;
    mat psi_desired;

    mat Td;
    mat Nd;
};

void PathFollowing::activate(const udrone::statusControlAlgorithms msg)
{
    if(msg.statusPathfollowing){
        this->pf_active = true;
        ROS_INFO("Switching to path following");
        this->pubConfirmChangeCtrlAlg.publish(this->returnMsg);
    } else {
        this->pf_active = false;
    }
}

void PathFollowing::updatePathCallback(const path_following::update_path& msg)
{
    this->S = 0;

    this->x_desired.reset();
    this->y_desired.reset();
    this->psi_desired.reset();
    this->Td.reset();
    this->Nd.reset();

    this->Nl = msg.Nl;
    this->kappa = msg.kappa;

    if (this->Nl <= 0) {
        this->Nl = 1;
        this->x_desired = zeros(this->Nl+1, 1);
        this->y_desired = zeros(this->Nl+1, 1);
        this->psi_desired = zeros(this->Nl+1, 1);
        
        this->Td = zeros(2, this->Nl+1);
        this->Nd = zeros(2, this->Nl+1);
    } else {
        this->x_desired = zeros(this->Nl, 1);
        this->y_desired = zeros(this->Nl, 1);
        this->psi_desired = zeros(this->Nl, 1);
        
        this->Td = zeros(2, this->Nl);
        this->Nd = zeros(2, this->Nl);
    }

    this->x_desired(0,0) = msg.initial_position_x;
    this->y_desired(0,0) = msg.initial_position_y;
    this->psi_desired(0,0) = msg.initial_heading;

    this->Td.col(0) = vec({cos(this->psi_desired(0,0)), sin(this->psi_desired(0,0))});
    this->Nd.col(0) = vec({-sin(this->psi_desired(0,0)), cos(this->psi_desired(0,0))});


    for (int i = 0; i < this->Nl-1; i++) {
        this->x_desired(i+1, 0) = this->x_desired(i) + this->Td(0,i) * this->ds;
        this->y_desired(i+1, 0) = this->y_desired(i) + this->Td(1,i) * this->ds;
        this->psi_desired(i+1, 0) = this->psi_desired(i) + this->kappa * this->ds;

        this->Td.col(i+1) = this->Td.col(i) + this->kappa * Nd.col(i) * ds;
        this->Td.col(i+1) = this->Td.col(i+1)/norm(this->Td.col(i+1));
        this->Nd.col(i+1) = this->Nd.col(i) - kappa * this->Td.col(i) * ds;
        this->Nd.col(i+1) = this->Nd.col(i+1)/norm(this->Nd.col(i+1));
    }

    this->current_z_desired = msg.depth;

    this->path_set = 1;
}


void PathFollowing::setVelocityCallback(const path_following::desired_velocity& msg)
{
    this->desired_velocity = msg.desired_velocity;
}

void PathFollowing::updateDesired(float loop_periode)
{
    if (this->path_set) {
        this->S += this->desired_velocity * loop_periode;

        float S_scaled = this->S/this->ds;

        int low_i = (int) S_scaled;
        int high_i = low_i + 1;

        if (low_i <= 0) {
            this->current_x_desired   = this->x_desired(0, 0);
            this->current_y_desired   = this->y_desired(0, 0);
            this->current_psi_desired = this->psi_desired(0, 0);

            if (this->desired_velocity <= 0) {
                this->S = 0;
            }
        } else if (this->x_desired.in_range(high_i, 0)) {
            this->current_x_desired   = this->lineareInterpolation(S_scaled, low_i, high_i, this->x_desired(low_i, 0), this->x_desired(high_i, 0));
            this->current_y_desired   = this->lineareInterpolation(S_scaled, low_i, high_i, this->y_desired(low_i, 0), this->y_desired(high_i, 0));
            this->current_psi_desired = this->lineareInterpolation(S_scaled, low_i, high_i, this->psi_desired(low_i, 0), this->psi_desired(high_i, 0));
        } else {
            int i = this->x_desired.n_rows - 1;

            this->current_x_desired   = this->x_desired(i, 0);
            this->current_y_desired   = this->y_desired(i, 0);
            this->current_psi_desired = this->psi_desired(i, 0);
        }
    }
}

float PathFollowing::lineareInterpolation(float x, float x0, float x1, float y0, float y1) {
    float y;

    y = y0 + ((x-x0)/(x1-x0)) * (y1-y0);

    return y;
}

int main(int argc, char **argv)
{
    float loop_periode = 0.1; // dt

    ros::init(argc, argv, "path_following");
    ros::NodeHandle n;

    PathFollowing PF(n);

    ros::Subscriber path_updater     = n.subscribe("/path_following/update_path", 100, &PathFollowing::updatePathCallback, &PF);
    ros::Subscriber velocity_updater = n.subscribe("/path_following/desired_velocity", 100, &PathFollowing::setVelocityCallback, &PF);

    ros::Publisher  path_publisher   = n.advertise<udrone::desiredPosition>("desired_position", 10);

    ros::Rate loop_rate(1/loop_periode); // Update frequency

    udrone::desiredPosition msg;

    while(ros::ok()) {
        if (PF.pf_active) {
            PF.updateDesired(loop_periode);

            msg.x_desired   = PF.current_x_desired;
            msg.y_desired   = PF.current_y_desired;
            msg.psi_desired = PF.current_psi_desired;
            msg.z_desired   = PF.current_z_desired;

            path_publisher.publish(msg);
        }

        ros::spinOnce();
        loop_rate.sleep();
    }

    return 0;
}
