#include <iostream>
#include "ros/ros.h"
#include <armadillo>
#include <math.h>

#include "geometry_msgs/Wrench.h"
#include "udrone/xhat.h"
#include "udrone/desiredPosition.h"
#include "udrone/forceCommand.h"
#include "udrone/statusControlAlgorithms.h"


using namespace arma;

class LQR
{
public:
    LQR();
    void spin();

private:
    ros::NodeHandle nh;

    // ROS subscriptions
    ros::Subscriber sub_xhat;
    ros::Subscriber sub_desired_position;
    ros::Subscriber subChangeCtrlAlg;

    // ROS publishers
    ros::Publisher pubConfirmChangeCtrlAlg;
    ros::Publisher pub_force_command;

    void updatePosition(const udrone::xhat);
    void updateDesired(const udrone::desiredPosition);
    void activate(const udrone::statusControlAlgorithms);
    void publishForce();
    void sendZeroForceCmd();

    void run();

    // System status variables
    bool controllerOn;
    udrone::statusControlAlgorithms returnMsg;
    float dt;

    // Controller matrices
    mat xhat;
    mat ref;
    mat u;
    mat P;
    mat J_psi_trans;
    mat K_c;

    udrone::forceCommand forceMessage;
};

LQR::LQR()
{
    pubConfirmChangeCtrlAlg     = nh.advertise<udrone::statusControlAlgorithms>("confirmChangeCtrlAlg",10);
    subChangeCtrlAlg            = nh.subscribe("changeCtrlAlg", 10, &LQR::activate, this);

    this->controllerOn                        = false;
    this->returnMsg.confirmationMsg           = true;
    this->returnMsg.statusDirectMotionControl = false;
    this->returnMsg.statusDirectThrustControl = false;
    this->returnMsg.statusAutoDepthControl    = false;
    this->returnMsg.statusAutoHeadingControl  = false;
    this->returnMsg.statusPathfollowing       = true;

    dt = 0.1;

    // Initialize matrices
    this->xhat        = zeros(12,1);
    this->ref         = zeros(4,1);
    this->P           = zeros(4,4);
    this->J_psi_trans = zeros(4,4);
    this->K_c         = zeros(4,8);


    // Initialize controller gain matrices
    this->P.col(0)    =    vec({2.1184,         0,         0,         0});
    this->P.col(1)    =    vec({     0,    2.1743,         0,         0});
    this->P.col(2)    =    vec({     0,         0,    6.2761,         0});
    this->P.col(3)    =    vec({     0,         0,         0,    0.0661});

    this->K_c.row(0)  = rowvec({2.1184,         0,         0,         0,    9.9301,         0,         0,         0});
    this->K_c.row(1)  = rowvec({     0,    2.1743,         0,         0,         0,    7.7424,         0,         0});
    this->K_c.row(2)  = rowvec({     0,         0,    6.2761,         0,         0,         0,   19.5089,         0});
    this->K_c.row(3)  = rowvec({     0,         0,         0,    0.0661,         0,         0,         0,    0.2779});

    // Initialize ros subscriptions
    sub_xhat             = nh.subscribe("xhat", 10, &LQR::updatePosition, this);
    sub_desired_position = nh.subscribe("desired_position", 10, &LQR::updateDesired, this);

    // Initialize ros publishings
    pub_force_command    = nh.advertise<udrone::forceCommand>("controlForce", 10);
}

void LQR::activate(const udrone::statusControlAlgorithms msg)
{
    if(msg.statusPathfollowing){
        this->sendZeroForceCmd();
        this->controllerOn = true;
        ROS_INFO("Switching to path following");
        this->pubConfirmChangeCtrlAlg.publish(this->returnMsg);
    } else {
        this->controllerOn = false;
    }
}

void LQR::sendZeroForceCmd()
{
    this->forceMessage.force.force.x   =    0;
    this->forceMessage.force.force.y    =   0;
    this->forceMessage.force.force.z    =   0;
    this->forceMessage.force.torque.z   =   0;
    this->forceMessage.logTime = ros::Time::now();
    this->pub_force_command.publish(this->forceMessage);
}

void LQR::updatePosition(const udrone::xhat _xhat)
{
    this->xhat(0,0) = _xhat.x;
    this->xhat(1,0) = _xhat.y;
    this->xhat(2,0) = _xhat.z;
    this->xhat(3,0) = _xhat.psi;
    this->xhat(4,0) = _xhat.u;
    this->xhat(5,0) = _xhat.v;
    this->xhat(6,0) = _xhat.w;
    this->xhat(7,0) = _xhat.r;
    this->xhat(8,0) = _xhat.b1;
    this->xhat(9,0) = _xhat.b2;
    this->xhat(10,0) = _xhat.b3;
    this->xhat(11,0) = _xhat.b4;
}

void LQR::updateDesired(const udrone::desiredPosition _desPos)
{
    this->ref(0,0) = _desPos.x_desired;
    this->ref(1,0) = _desPos.y_desired;
    this->ref(2,0) = _desPos.z_desired;
    this->ref(3,0) = _desPos.psi_desired;
}

// Publish force command to ros with saturation limits
void LQR::publishForce()
{
    this->forceMessage.force.force.x = this->u(0,0);
    this->forceMessage.force.force.y = this->u(1,0);
    this->forceMessage.force.force.z = this->u(2,0);
    this->forceMessage.force.torque.x = 0;
    this->forceMessage.force.torque.y = 0;
    this->forceMessage.force.torque.z = this->u(3,0);
    
    this->pub_force_command.publish(this->forceMessage);
}

// Calculate the desired thrust output
void LQR::run()
{
    float psi = xhat(3,0);
    this->J_psi_trans.row(0) = rowvec({ cos(psi),  sin(psi), 0, 0});
    this->J_psi_trans.row(1) = rowvec({-sin(psi),  cos(psi), 0, 0});
    this->J_psi_trans.row(2) = rowvec({   0,          0,     1, 0});
    this->J_psi_trans.row(3) = rowvec({   0,          0,     0, 1});

    this->u = this->P * this->ref - this->K_c * xhat(span(0,7), 0) - this->J_psi_trans * this->xhat(span(8,11), 0);

    this->u(0,0) = fmax(fmin(this->u(0,0), 0.3), -0.3);
    this->u(1,0) = fmax(fmin(this->u(1,0), 0.7), -0.7);
    this->u(2,0) = fmax(fmin(this->u(2,0), 2.0), -2.0);
    this->u(3,0) = fmax(fmin(this->u(3,0), 0.8), -0.8);
}

void LQR::spin()
{
    ros::Rate loop(1/this->dt);
    while(ros::ok()) {
        if (this->controllerOn) {
            this->run();
            this->publishForce();
        }
        loop.sleep();
        ros::spinOnce();
    }
}

int main(int argc, char **argv)
{
    ros::init(argc, argv, "LQR");
    LQR controller;
    controller.spin();
    return 0;
}